@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -38,8 +38,6 @@
 #include "simutil.h"
 #include "storage.h"
 
-#define CBS_MANAGER_INTERFACE "org.ofono.CbsManager"
-
 #define SETTINGS_STORE "cbs"
 #define SETTINGS_GROUP "Settings"
 
@@ -59,6 +57,7 @@ struct ofono_cbs {
 	GSList *topics;
 	GSList *new_topics;
 	struct ofono_sim *sim;
+	struct ofono_stk *stk;
 	struct ofono_netreg *netreg;
 	unsigned int netreg_watch;
 	unsigned int location_watch;
@@ -130,7 +129,7 @@ static void cbs_dispatch_emergency(struct ofono_cbs *cbs, const char *message,
 		return;
 	};
 
-	signal = dbus_message_new_signal(path, CBS_MANAGER_INTERFACE,
+	signal = dbus_message_new_signal(path, OFONO_CELL_BROADCAST_INTERFACE,
 						"EmergencyBroadcast");
 
 	if (!signal)
@@ -164,7 +163,7 @@ static void cbs_dispatch_text(struct ofono_cbs *cbs, enum sms_class cls,
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path = __ofono_atom_get_path(cbs->atom);
 
-	g_dbus_emit_signal(conn, path, CBS_MANAGER_INTERFACE,
+	g_dbus_emit_signal(conn, path, OFONO_CELL_BROADCAST_INTERFACE,
 				"IncomingBroadcast",
 				DBUS_TYPE_STRING, &message,
 				DBUS_TYPE_UINT16, &channel,
@@ -186,16 +185,19 @@ void ofono_cbs_notify(struct ofono_cbs *cbs, const unsigned char *pdu,
 	if (cbs->assembly == NULL)
 		return;
 
-	if (!cbs->powered)
-		return;
-
 	if (!cbs_decode(pdu, pdu_len, &c)) {
 		ofono_error("Unable to decode CBS PDU");
 		return;
 	}
 
 	if (cbs_topic_in_range(c.message_identifier, cbs->efcbmid_contents)) {
-		__ofono_cbs_sim_download(cbs->sim, pdu, pdu_len);
+		if (cbs->stk)
+			__ofono_cbs_sim_download(cbs->stk, &c);
+		return;
+	}
+
+	if (!cbs->powered) {
+		ofono_error("Ignoring CBS because powered is off");
 		return;
 	}
 
@@ -300,8 +302,6 @@ static char *cbs_topics_to_str(struct ofono_cbs *cbs, GSList *user_topics)
 	char *topic_str;
 	struct cbs_topic_range etws_range = { 4352, 4356 };
 
-	topics = g_slist_append(topics, &etws_range);
-
 	if (user_topics != NULL)
 		topics = g_slist_concat(topics,
 					g_slist_copy(user_topics));
@@ -310,6 +310,8 @@ static char *cbs_topics_to_str(struct ofono_cbs *cbs, GSList *user_topics)
 		topics = g_slist_concat(topics,
 					g_slist_copy(cbs->efcbmid_contents));
 
+	topics = g_slist_append(topics, &etws_range);
+
 	topic_str = cbs_topic_ranges_to_string(topics);
 	g_slist_free(topics);
 
@@ -329,7 +331,7 @@ static void cbs_set_topics_cb(const struct ofono_error *error, void *data)
 		g_slist_free(cbs->new_topics);
 		cbs->new_topics = NULL;
 
-		ofono_debug("Setting Cell Broadcast topics failed");
+		DBG("Setting Cell Broadcast topics failed");
 		__ofono_dbus_pending_reply(&cbs->pending,
 					__ofono_error_failed(cbs->pending));
 		return;
@@ -345,9 +347,15 @@ static void cbs_set_topics_cb(const struct ofono_error *error, void *data)
 
 	topics = cbs_topic_ranges_to_string(cbs->topics);
 	ofono_dbus_signal_property_changed(conn, path,
-						CBS_MANAGER_INTERFACE,
+						OFONO_CELL_BROADCAST_INTERFACE,
 						"Topics",
 						DBUS_TYPE_STRING, &topics);
+
+	if (cbs->settings) {
+		g_key_file_set_string(cbs->settings, SETTINGS_GROUP,
+					"Topics", topics);
+		storage_sync(cbs->imsi, SETTINGS_STORE, cbs->settings);
+	}
 	g_free(topics);
 }
 
@@ -391,7 +399,8 @@ static void cbs_set_powered_cb(const struct ofono_error *error, void *data)
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Setting Cell Broadcast topics failed");
+		ofono_error("Setting Cell Broadcast topics failed");
+
 		if (!cbs->pending)
 			return;
 
@@ -409,7 +418,7 @@ static void cbs_set_powered_cb(const struct ofono_error *error, void *data)
 	}
 
 	ofono_dbus_signal_property_changed(conn, path,
-						CBS_MANAGER_INTERFACE,
+						OFONO_CELL_BROADCAST_INTERFACE,
 						"Powered",
 						DBUS_TYPE_BOOLEAN,
 						&cbs->powered);
@@ -431,7 +440,7 @@ static DBusMessage *cbs_set_powered(struct ofono_cbs *cbs, gboolean value,
 	if (cbs->powered == value)
 		goto reply;
 
-	if (!cbs->driver->set_topics)
+	if (!cbs->driver->set_topics || !cbs->driver->clear_topics)
 		goto done;
 
 	if (msg)
@@ -458,7 +467,7 @@ done:
 	}
 
 	ofono_dbus_signal_property_changed(conn, path,
-						CBS_MANAGER_INTERFACE,
+						OFONO_CELL_BROADCAST_INTERFACE,
 						"Powered",
 						DBUS_TYPE_BOOLEAN,
 						&cbs->powered);
@@ -520,14 +529,14 @@ static DBusMessage *cbs_set_property(DBusConnection *conn, DBusMessage *msg,
 	return __ofono_error_invalid_args(msg);
 }
 
-static GDBusMethodTable cbs_manager_methods[] = {
+static GDBusMethodTable cbs_methods[] = {
 	{ "GetProperties",	"",	"a{sv}",	cbs_get_properties },
 	{ "SetProperty",	"sv",	"",		cbs_set_property,
 							G_DBUS_METHOD_FLAG_ASYNC },
 	{ }
 };
 
-static GDBusSignalTable cbs_manager_signals[] = {
+static GDBusSignalTable cbs_signals[] = {
 	{ "PropertyChanged",	"sv"		},
 	{ "IncomingBroadcast",	"sq"		},
 	{ "EmergencyBroadcast", "sa{sv}"	},
@@ -560,8 +569,8 @@ static void cbs_unregister(struct ofono_atom *atom)
 	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
 	const char *path = __ofono_atom_get_path(atom);
 
-	g_dbus_unregister_interface(conn, path, CBS_MANAGER_INTERFACE);
-	ofono_modem_remove_interface(modem, CBS_MANAGER_INTERFACE);
+	g_dbus_unregister_interface(conn, path, OFONO_CELL_BROADCAST_INTERFACE);
+	ofono_modem_remove_interface(modem, OFONO_CELL_BROADCAST_INTERFACE);
 
 	if (cbs->topics) {
 		g_slist_foreach(cbs->topics, (GFunc)g_free, NULL);
@@ -583,6 +592,7 @@ static void cbs_unregister(struct ofono_atom *atom)
 	}
 
 	cbs->sim = NULL;
+	cbs->stk = NULL;
 
 	if (cbs->reset_source) {
 		g_source_remove(cbs->reset_source);
@@ -669,6 +679,58 @@ struct ofono_cbs *ofono_cbs_create(struct ofono_modem *modem,
 	return cbs;
 }
 
+static void cbs_got_file_contents(struct ofono_cbs *cbs)
+{
+	gboolean powered;
+	GSList *initial_topics = NULL;
+	char *topics_str;
+	GError *error = NULL;
+
+	if (cbs->topics == NULL) {
+		if (cbs->efcbmi_contents != NULL)
+			initial_topics = g_slist_concat(initial_topics,
+					g_slist_copy(cbs->efcbmi_contents));
+
+		if (cbs->efcbmir_contents != NULL)
+			initial_topics = g_slist_concat(initial_topics,
+					g_slist_copy(cbs->efcbmir_contents));
+
+		cbs->topics = cbs_optimize_ranges(initial_topics);
+		g_slist_free(initial_topics);
+
+		topics_str = cbs_topic_ranges_to_string(cbs->topics);
+		g_key_file_set_string(cbs->settings, SETTINGS_GROUP,
+					"Topics", topics_str);
+		g_free(topics_str);
+		storage_sync(cbs->imsi, SETTINGS_STORE, cbs->settings);
+	}
+
+	if (cbs->efcbmi_length) {
+		cbs->efcbmi_length = 0;
+		g_slist_foreach(cbs->efcbmi_contents, (GFunc) g_free, NULL);
+		g_slist_free(cbs->efcbmi_contents);
+		cbs->efcbmi_contents = NULL;
+	}
+
+	if (cbs->efcbmir_length) {
+		cbs->efcbmir_length = 0;
+		g_slist_foreach(cbs->efcbmir_contents, (GFunc) g_free, NULL);
+		g_slist_free(cbs->efcbmir_contents);
+		cbs->efcbmir_contents = NULL;
+	}
+
+	powered = g_key_file_get_boolean(cbs->settings, SETTINGS_GROUP,
+						"Powered", &error);
+
+	if (error) {
+		powered = TRUE;
+		g_key_file_set_boolean(cbs->settings, SETTINGS_GROUP,
+					"Powered", powered);
+	}
+
+	cbs_set_powered(cbs, powered, NULL);
+}
+
 static void sim_cbmi_read_cb(int ok, int length, int record,
 				const unsigned char *data,
 				int record_length, void *userdata)
@@ -708,16 +770,9 @@ static void sim_cbmi_read_cb(int ok, int length, int record,
 	if (cbs->efcbmi_contents == NULL)
 		return;
 
-	cbs->efcbmi_contents = g_slist_reverse(cbs->efcbmi_contents);
-
 	str = cbs_topic_ranges_to_string(cbs->efcbmi_contents);
-	ofono_debug("Got cbmi: %s", str);
+	DBG("Got cbmi: %s", str);
 	g_free(str);
-
-	cbs->efcbmi_length = 0;
-	g_slist_foreach(cbs->efcbmi_contents, (GFunc)g_free, NULL);
-	g_slist_free(cbs->efcbmi_contents);
-	cbs->efcbmi_contents = NULL;
 }
 
 static void sim_cbmir_read_cb(int ok, int length, int record,
@@ -762,16 +817,9 @@ static void sim_cbmir_read_cb(int ok, int length, int record,
 	if (cbs->efcbmir_contents == NULL)
 		return;
 
-	cbs->efcbmir_contents = g_slist_reverse(cbs->efcbmir_contents);
-
 	str = cbs_topic_ranges_to_string(cbs->efcbmir_contents);
-	ofono_debug("Got cbmir: %s", str);
+	DBG("Got cbmir: %s", str);
 	g_free(str);
-
-	cbs->efcbmir_length = 0;
-	g_slist_foreach(cbs->efcbmir_contents, (GFunc)g_free, NULL);
-	g_slist_free(cbs->efcbmir_contents);
-	cbs->efcbmir_contents = NULL;
 }
 
 static void sim_cbmid_read_cb(int ok, int length, int record,
@@ -785,10 +833,10 @@ static void sim_cbmid_read_cb(int ok, int length, int record,
 	GSList *contents = NULL;
 
 	if (!ok)
-		return;
+		goto done;
 
 	if ((length % 2) == 1 || length < 2)
-		return;
+		goto done;
 
 	cbs->efcbmid_length = length;
 
@@ -808,21 +856,24 @@ static void sim_cbmid_read_cb(int ok, int length, int record,
 	}
 
 	if (contents == NULL)
-		return;
+		goto done;
 
 	cbs->efcbmid_contents = g_slist_reverse(contents);
 
 	str = cbs_topic_ranges_to_string(cbs->efcbmid_contents);
-	ofono_debug("Got cbmid: %s", str);
+	DBG("Got cbmid: %s", str);
 	g_free(str);
+
+done:
+	cbs_got_file_contents(cbs);
 }
 
 static void cbs_got_imsi(struct ofono_cbs *cbs)
 {
 	const char *imsi = ofono_sim_get_imsi(cbs->sim);
-	gboolean powered;
+	char *topics_str;
 
-	ofono_debug("Got IMSI: %s", imsi);
+	DBG("Got IMSI: %s", imsi);
 
 	cbs->settings = storage_open(imsi, SETTINGS_STORE);
 	if (cbs->settings == NULL)
@@ -830,16 +881,28 @@ static void cbs_got_imsi(struct ofono_cbs *cbs)
 
 	cbs->imsi = g_strdup(imsi);
 
-	powered = g_key_file_get_boolean(cbs->settings, SETTINGS_GROUP,
-						"Powered", NULL);
-	cbs_set_powered(cbs, powered, NULL);
+	cbs->topics = NULL;
+
+	topics_str = g_key_file_get_string(cbs->settings, SETTINGS_GROUP,
+						"Topics", NULL);
+	if (topics_str)
+		cbs->topics = cbs_extract_topic_ranges(topics_str);
+
+	/* If stored value is invalid or no stored value, bootstrap
+	 * topics list from SIM contents */
+	if (topics_str == NULL ||
+			(cbs->topics == NULL && topics_str[0] != '\0')) {
+		ofono_sim_read(cbs->sim, SIM_EFCBMI_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+				sim_cbmi_read_cb, cbs);
+		ofono_sim_read(cbs->sim, SIM_EFCBMIR_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+				sim_cbmir_read_cb, cbs);
+	}
+
+	if (topics_str)
+		g_free(topics_str);
 
-	ofono_sim_read(cbs->sim, SIM_EFCBMI_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_cbmi_read_cb, cbs);
-	ofono_sim_read(cbs->sim, SIM_EFCBMIR_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_cbmir_read_cb, cbs);
 	ofono_sim_read(cbs->sim, SIM_EFCBMID_FILEID,
 			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
 			sim_cbmid_read_cb, cbs);
@@ -915,7 +978,7 @@ out:
 	 * that time
 	 */
 	if (lac_changed || ci_changed) {
-		cbs->reset_source = 
+		cbs->reset_source =
 			g_timeout_add_seconds(3, reset_base_station_name, cbs);
 	}
 
@@ -967,29 +1030,34 @@ void ofono_cbs_register(struct ofono_cbs *cbs)
 	struct ofono_modem *modem = __ofono_atom_get_modem(cbs->atom);
 	const char *path = __ofono_atom_get_path(cbs->atom);
 	struct ofono_atom *sim_atom;
+	struct ofono_atom *stk_atom;
 	struct ofono_atom *netreg_atom;
 
 	if (!g_dbus_register_interface(conn, path,
-					CBS_MANAGER_INTERFACE,
-					cbs_manager_methods,
-					cbs_manager_signals,
+					OFONO_CELL_BROADCAST_INTERFACE,
+					cbs_methods, cbs_signals,
 					NULL, cbs, NULL)) {
 		ofono_error("Could not create %s interface",
-				CBS_MANAGER_INTERFACE);
+				OFONO_CELL_BROADCAST_INTERFACE);
 		return;
 	}
 
-	ofono_modem_add_interface(modem, CBS_MANAGER_INTERFACE);
+	ofono_modem_add_interface(modem, OFONO_CELL_BROADCAST_INTERFACE);
 
 	sim_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SIM);
 
 	if (sim_atom) {
 		cbs->sim = __ofono_atom_get_data(sim_atom);
 
-		if (ofono_sim_get_ready(cbs->sim) == TRUE)
+		if (ofono_sim_get_state(cbs->sim) == OFONO_SIM_STATE_READY)
 			cbs_got_imsi(cbs);
 	}
 
+	stk_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_STK);
+
+	if (stk_atom)
+		cbs->stk = __ofono_atom_get_data(stk_atom);
+
 	cbs->netreg_watch = __ofono_modem_add_atom_watch(modem,
 					OFONO_ATOM_TYPE_NETREG,
 					netreg_watch, cbs, NULL);
