@@ -776,6 +776,20 @@ static DBusMessage *network_get_properties(DBusConnection *conn,
 					&technology);
 	}
 
+	if (netreg->current_operator) {
+		if (netreg->current_operator->mcc[0] != '\0') {
+			const char *mcc = netreg->current_operator->mcc;
+			ofono_dbus_dict_append(&dict, "MobileCountryCode",
+						DBUS_TYPE_STRING, &mcc);
+		}
+
+		if (netreg->current_operator->mnc[0] != '\0') {
+			const char *mnc = netreg->current_operator->mnc;
+			ofono_dbus_dict_append(&dict, "MobileNetworkCode",
+						DBUS_TYPE_STRING, &mnc);
+		}
+	}
+
 	operator = get_operator_display_name(netreg);
 	ofono_dbus_dict_append(&dict, "Name", DBUS_TYPE_STRING, &operator);
 
@@ -1107,7 +1121,7 @@ static GSList *compress_operator_list(const struct ofono_network_operator *list,
 }
 
 static gboolean update_operator_list(struct ofono_netreg *netreg, int total,
-					const struct ofono_network_operator *list)
+				const struct ofono_network_operator *list)
 {
 	GSList *n = NULL;
 	GSList *o;
@@ -1168,8 +1182,8 @@ static gboolean update_operator_list(struct ofono_netreg *netreg, int total,
 }
 
 static void operator_list_callback(const struct ofono_error *error, int total,
-					const struct ofono_network_operator *list,
-					void *data)
+				const struct ofono_network_operator *list,
+				void *data)
 {
 	struct ofono_netreg *netreg = data;
 	DBusMessage *reply;
@@ -1220,6 +1234,15 @@ static void current_operator_callback(const struct ofono_error *error,
 
 	DBG("%p, %p", netreg, netreg->current_operator);
 
+	/*
+	 * Sometimes we try to query COPS right when we roam off the cell,
+	 * in which case the operator information frequently comes in bogus.
+	 * We ignore it here
+	 */
+	if (netreg->status != NETWORK_REGISTRATION_STATUS_REGISTERED &&
+			netreg->status != NETWORK_REGISTRATION_STATUS_ROAMING)
+		current = NULL;
+
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		DBG("Error during current operator");
 		return;
@@ -1290,6 +1313,24 @@ emit:
 					"Name", DBUS_TYPE_STRING,
 					&operator);
 
+	if (netreg->current_operator) {
+		if (netreg->current_operator->mcc[0] != '\0') {
+			const char *mcc = netreg->current_operator->mcc;
+			ofono_dbus_signal_property_changed(conn, path,
+					OFONO_NETWORK_REGISTRATION_INTERFACE,
+					"MobileCountryCode",
+					DBUS_TYPE_STRING, &mcc);
+		}
+
+		if (netreg->current_operator->mnc[0] != '\0') {
+			const char *mnc = netreg->current_operator->mnc;
+			ofono_dbus_signal_property_changed(conn, path,
+					OFONO_NETWORK_REGISTRATION_INTERFACE,
+					"MobileNetworkCode",
+					DBUS_TYPE_STRING, &mnc);
+		}
+	}
+
 	notify_status_watches(netreg);
 }
 
