@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -37,9 +37,6 @@
 #include "simutil.h"
 #include "smsutil.h"
 
-#define VOICECALL_MANAGER_INTERFACE "org.ofono.VoiceCallManager"
-#define VOICECALL_INTERFACE "org.ofono.VoiceCall"
-
 #define VOICECALLS_FLAG_MULTI_RELEASE 0x1
 
 #define MAX_VOICE_CALLS 16
@@ -56,7 +53,9 @@ struct ofono_voicecall {
 	DBusMessage *pending;
 	gint emit_calls_source;
 	gint emit_multi_source;
+	struct ofono_sim *sim;
 	unsigned int sim_watch;
+	unsigned int sim_state_watch;
 	const struct ofono_voicecall_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
@@ -176,7 +175,7 @@ static DBusMessage *voicecall_get_properties(DBusConnection *conn,
 	DBusMessageIter dict;
 	const char *status;
 	const char *callerid;
-	const char *timestr = "";
+	const char *timestr;
 
 	reply = dbus_message_new_method_return(msg);
 
@@ -198,8 +197,9 @@ static DBusMessage *voicecall_get_properties(DBusConnection *conn,
 				DBUS_TYPE_STRING, &callerid);
 
 	if (call->status == CALL_STATUS_ACTIVE ||
-		(call->status == CALL_STATUS_DISCONNECTED && v->start_time != 0) ||
-		call->status == CALL_STATUS_HELD) {
+			call->status == CALL_STATUS_HELD ||
+			(call->status == CALL_STATUS_DISCONNECTED &&
+				v->start_time != 0)) {
 		timestr = time_to_str(&v->start_time);
 
 		ofono_dbus_dict_append(&dict, "StartTime", DBUS_TYPE_STRING,
@@ -400,7 +400,8 @@ static void voicecall_emit_disconnect_reason(struct voicecall *call,
 	reason_str = disconnect_reason_to_string(reason);
 	path = voicecall_build_path(call->vc, call->call);
 
-	g_dbus_emit_signal(conn, path, VOICECALL_INTERFACE, "DisconnectReason",
+	g_dbus_emit_signal(conn, path, OFONO_VOICECALL_INTERFACE,
+				"DisconnectReason",
 				DBUS_TYPE_STRING, &reason_str,
 				DBUS_TYPE_INVALID);
 }
@@ -423,7 +424,8 @@ static void voicecall_set_call_status(struct voicecall *call,
 	status_str = call_status_to_string(status);
 	path = voicecall_build_path(call->vc, call->call);
 
-	ofono_dbus_signal_property_changed(conn, path, VOICECALL_INTERFACE,
+	ofono_dbus_signal_property_changed(conn, path,
+						OFONO_VOICECALL_INTERFACE,
 						"State", DBUS_TYPE_STRING,
 						&status_str);
 
@@ -438,10 +440,9 @@ static void voicecall_set_call_status(struct voicecall *call,
 		timestr = time_to_str(&call->start_time);
 
 		ofono_dbus_signal_property_changed(conn, path,
-							VOICECALL_INTERFACE,
-							"StartTime",
-							DBUS_TYPE_STRING,
-							&timestr);
+						OFONO_VOICECALL_INTERFACE,
+						"StartTime", DBUS_TYPE_STRING,
+						&timestr);
 	}
 }
 
@@ -480,7 +481,8 @@ static void voicecall_set_call_lineid(struct voicecall *v,
 	else
 		lineid_str = phone_number_to_string(ph);
 
-	ofono_dbus_signal_property_changed(conn, path, VOICECALL_INTERFACE,
+	ofono_dbus_signal_property_changed(conn, path,
+						OFONO_VOICECALL_INTERFACE,
 						"LineIdentification",
 						DBUS_TYPE_STRING, &lineid_str);
 }
@@ -495,7 +497,7 @@ static gboolean voicecall_dbus_register(struct voicecall *v)
 
 	path = voicecall_build_path(v->vc, v->call);
 
-	if (!g_dbus_register_interface(conn, path, VOICECALL_INTERFACE,
+	if (!g_dbus_register_interface(conn, path, OFONO_VOICECALL_INTERFACE,
 					voicecall_methods,
 					voicecall_signals,
 					NULL, v, voicecall_destroy)) {
@@ -515,7 +517,7 @@ static gboolean voicecall_dbus_unregister(struct ofono_voicecall *vc,
 	const char *path = voicecall_build_path(vc, v->call);
 
 	return g_dbus_unregister_interface(conn, path,
-					VOICECALL_INTERFACE);
+						OFONO_VOICECALL_INTERFACE);
 }
 
 
@@ -692,10 +694,9 @@ static gboolean real_emit_call_list_changed(void *data)
 	voicecalls_path_list(vc, vc->call_list, &objpath_list);
 
 	ofono_dbus_signal_array_property_changed(conn, path,
-				VOICECALL_MANAGER_INTERFACE,
-				"Calls",
-				DBUS_TYPE_OBJECT_PATH,
-				&objpath_list);
+					OFONO_VOICECALL_MANAGER_INTERFACE,
+					"Calls", DBUS_TYPE_OBJECT_PATH,
+					&objpath_list);
 
 	g_strfreev(objpath_list);
 
@@ -721,9 +722,9 @@ static gboolean real_emit_multiparty_call_list_changed(void *data)
 	voicecalls_path_list(vc, vc->multiparty_list, &objpath_list);
 
 	ofono_dbus_signal_array_property_changed(conn, path,
-				VOICECALL_MANAGER_INTERFACE, "MultipartyCalls",
-				DBUS_TYPE_OBJECT_PATH,
-				&objpath_list);
+					OFONO_VOICECALL_MANAGER_INTERFACE,
+					"MultipartyCalls",
+					DBUS_TYPE_OBJECT_PATH, &objpath_list);
 
 	g_strfreev(objpath_list);
 
@@ -829,8 +830,9 @@ static ofono_bool_t clir_string_to_clir(const char *clirstr,
 	} else if (!strcmp(clirstr, "enabled")) {
 		*clir = OFONO_CLIR_OPTION_INVOCATION;
 		return TRUE;
-	} else
+	} else {
 		return FALSE;
+	}
 }
 
 static struct ofono_call *synthesize_outgoing_call(struct ofono_voicecall *vc,
@@ -878,7 +880,7 @@ static void dial_callback(const struct ofono_error *error, void *data)
 	gboolean need_to_emit = FALSE;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Dial callback returned error: %s",
+		DBG("Dial callback returned error: %s",
 			telephony_error_to_str(error));
 		reply = __ofono_error_failed(vc->pending);
 		__ofono_dbus_pending_reply(&vc->pending, reply);
@@ -916,7 +918,7 @@ static void dial_callback(const struct ofono_error *error, void *data)
 
 		v->detect_time = time(NULL);
 
-		ofono_debug("Registering new call: %d", call->id);
+		DBG("Registering new call: %d", call->id);
 		voicecall_dbus_register(v);
 
 		vc->call_list = g_slist_insert_sorted(vc->call_list, v,
@@ -1166,7 +1168,7 @@ static void private_chat_callback(const struct ofono_error *error, void *data)
 	GSList *l;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("command failed with error: %s",
+		DBG("command failed with error: %s",
 				telephony_error_to_str(error));
 		__ofono_dbus_pending_reply(&vc->pending,
 					__ofono_error_failed(vc->pending));
@@ -1260,7 +1262,7 @@ static void multiparty_create_callback(const struct ofono_error *error, void *da
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("command failed with error: %s",
+		DBG("command failed with error: %s",
 				telephony_error_to_str(error));
 		__ofono_dbus_pending_reply(&vc->pending,
 					__ofono_error_failed(vc->pending));
@@ -1464,7 +1466,7 @@ void ofono_voicecall_disconnected(struct ofono_voicecall *vc, int id,
 	time_t ts;
 	enum call_status prev_status;
 
-	ofono_debug("Got disconnection event for id: %d, reason: %d", id, reason);
+	DBG("Got disconnection event for id: %d, reason: %d", id, reason);
 
 	__ofono_modem_callid_release(modem, id);
 
@@ -1526,14 +1528,14 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 	struct voicecall *v = NULL;
 	struct ofono_call *newcall;
 
-	ofono_debug("Got a voicecall event, status: %d, id: %u, number: %s",
+	DBG("Got a voicecall event, status: %d, id: %u, number: %s",
 			call->status, call->id, call->phone_number.number);
 
 	l = g_slist_find_custom(vc->call_list, GUINT_TO_POINTER(call->id),
 				call_compare_by_id);
 
 	if (l) {
-		ofono_debug("Found call with id: %d\n", call->id);
+		DBG("Found call with id: %d\n", call->id);
 		voicecall_set_call_status(l->data, call->status);
 		voicecall_set_call_lineid(l->data, &call->phone_number,
 						call->clip_validity);
@@ -1541,7 +1543,7 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 		return;
 	}
 
-	ofono_debug("Did not find a call with id: %d\n", call->id);
+	DBG("Did not find a call with id: %d\n", call->id);
 
 	__ofono_modem_callid_hold(modem, call->id);
 
@@ -1549,21 +1551,21 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 
 	if (!newcall) {
 		ofono_error("Unable to allocate call");
-		goto err;
+		goto error;
 	}
 
 	v = voicecall_create(vc, newcall);
 
 	if (!v) {
 		ofono_error("Unable to allocate voicecall_data");
-		goto err;
+		goto error;
 	}
 
 	v->detect_time = time(NULL);
 
 	if (!voicecall_dbus_register(v)) {
 		ofono_error("Unable to register voice call");
-		goto err;
+		goto error;
 	}
 
 	vc->call_list = g_slist_insert_sorted(vc->call_list, v, call_compare);
@@ -1572,7 +1574,7 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 
 	return;
 
-err:
+error:
 	if (newcall)
 		g_free(newcall);
 
@@ -1586,7 +1588,7 @@ static void generic_callback(const struct ofono_error *error, void *data)
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
-		ofono_debug("command failed with error: %s",
+		DBG("command failed with error: %s",
 				telephony_error_to_str(error));
 
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR)
@@ -1626,10 +1628,8 @@ static void emit_en_list_changed(struct ofono_voicecall *vc)
 		list[i] = g_strdup(l->data);
 
 	ofono_dbus_signal_array_property_changed(conn, path,
-				VOICECALL_MANAGER_INTERFACE,
-				"EmergencyNumbers",
-				DBUS_TYPE_STRING,
-				&list);
+				OFONO_VOICECALL_MANAGER_INTERFACE,
+				"EmergencyNumbers", DBUS_TYPE_STRING, &list);
 
 	g_strfreev(list);
 }
@@ -1663,8 +1663,42 @@ static void set_new_ecc(struct ofono_voicecall *vc)
 	emit_en_list_changed(vc);
 }
 
-static void ecc_read_cb(int ok, int total_length, int record, const unsigned char *data,
-			int record_length, void *userdata)
+static void ecc_g2_read_cb(int ok, int total_length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_voicecall *vc = userdata;
+	char en[7];
+
+	DBG("%d", ok);
+
+	if (!ok)
+		return;
+
+	if (total_length < 3) {
+		ofono_error("Unable to read emergency numbers from SIM");
+		return;
+	}
+
+	total_length /= 3;
+	while (total_length--) {
+		extract_bcd_number(data, 3, en);
+		data += 3;
+
+		if (en[0] != '\0')
+			vc->new_en_list = g_slist_prepend(vc->new_en_list,
+								g_strdup(en));
+	}
+
+	if (vc->new_en_list == NULL)
+		return;
+
+	set_new_ecc(vc);
+}
+
+static void ecc_g3_read_cb(int ok, int total_length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
 {
 	struct ofono_voicecall *vc = userdata;
 	int total;
@@ -1744,9 +1778,9 @@ static void voicecall_unregister(struct ofono_atom *atom)
 
 	g_slist_free(vc->call_list);
 
-	ofono_modem_remove_interface(modem, VOICECALL_MANAGER_INTERFACE);
+	ofono_modem_remove_interface(modem, OFONO_VOICECALL_MANAGER_INTERFACE);
 	g_dbus_unregister_interface(conn, path,
-					VOICECALL_MANAGER_INTERFACE);
+					OFONO_VOICECALL_MANAGER_INTERFACE);
 }
 
 static void voicecall_remove(struct ofono_atom *atom)
@@ -1773,6 +1807,12 @@ static void voicecall_remove(struct ofono_atom *atom)
 		vc->new_en_list = NULL;
 	}
 
+	if (vc->sim_state_watch) {
+		ofono_sim_remove_state_watch(vc->sim, vc->sim_state_watch);
+		vc->sim_state_watch = 0;
+		vc->sim = NULL;
+	}
+
 	g_free(vc);
 }
 
@@ -1811,6 +1851,40 @@ struct ofono_voicecall *ofono_voicecall_create(struct ofono_modem *modem,
 	return vc;
 }
 
+static void sim_state_watch(void *user, enum ofono_sim_state new_state)
+{
+	struct ofono_voicecall *vc = user;
+
+	switch (new_state) {
+	case OFONO_SIM_STATE_INSERTED:
+		/* Try both formats, only one or none will work */
+		ofono_sim_read(vc->sim, SIM_EFECC_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+				ecc_g2_read_cb, vc);
+		ofono_sim_read(vc->sim, SIM_EFECC_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_FIXED,
+				ecc_g3_read_cb, vc);
+		break;
+	case OFONO_SIM_STATE_NOT_PRESENT:
+		/* TODO: Must release all non-emergency calls */
+
+		/*
+		 * Free the currently being read EN list, just in case the
+		 * SIM is removed when we're still reading them
+		 */
+		if (vc->new_en_list) {
+			g_slist_foreach(vc->new_en_list, (GFunc) g_free, NULL);
+			g_slist_free(vc->new_en_list);
+			vc->new_en_list = NULL;
+		}
+
+		add_to_en_list(&vc->new_en_list, default_en_list_no_sim);
+		set_new_ecc(vc);
+	default:
+		break;
+	}
+}
+
 static void sim_watch(struct ofono_atom *atom,
 			enum ofono_atom_watch_condition cond, void *data)
 {
@@ -1818,11 +1892,17 @@ static void sim_watch(struct ofono_atom *atom,
 	struct ofono_sim *sim = __ofono_atom_get_data(atom);
 
 	if (cond == OFONO_ATOM_WATCH_CONDITION_UNREGISTERED) {
+		vc->sim_state_watch = 0;
+		vc->sim = NULL;
 		return;
 	}
 
-	ofono_sim_read(sim, SIM_EFECC_FILEID, OFONO_SIM_FILE_STRUCTURE_FIXED,
-			ecc_read_cb, vc);
+	vc->sim = sim;
+	vc->sim_state_watch = ofono_sim_add_state_watch(sim,
+							sim_state_watch,
+							vc, NULL);
+
+	sim_state_watch(vc, ofono_sim_get_state(sim));
 }
 
 void ofono_voicecall_register(struct ofono_voicecall *vc)
@@ -1833,16 +1913,16 @@ void ofono_voicecall_register(struct ofono_voicecall *vc)
 	struct ofono_atom *sim_atom;
 
 	if (!g_dbus_register_interface(conn, path,
-					VOICECALL_MANAGER_INTERFACE,
+					OFONO_VOICECALL_MANAGER_INTERFACE,
 					manager_methods, manager_signals, NULL,
 					vc, NULL)) {
 		ofono_error("Could not create %s interface",
-				VOICECALL_MANAGER_INTERFACE);
+				OFONO_VOICECALL_MANAGER_INTERFACE);
 
 		return;
 	}
 
-	ofono_modem_add_interface(modem, VOICECALL_MANAGER_INTERFACE);
+	ofono_modem_add_interface(modem, OFONO_VOICECALL_MANAGER_INTERFACE);
 
 	/* Start out with the 22.101 mandated numbers, if we have a SIM and
 	 * the SIM contains EFecc, then we update the list once we've read them
